#ifndef XSECTION_H_
#define XSECTION_H_

#include <vector>
#include <string>
#include <complex>
#include "tred.h"
#include "coefficient.h"
#include "poly3.h"

struct XsectCmplxTerm {
  double coeff;
  int ssind;
  int ismeind;
};

struct XsectPolyTerm {
  double coeff;
  int ssind;
  int ismeind;
  int polyind;
};

std::vector<std::vector<std::pair<int, int> > > get_isme_sums(
    const std::string & filename);
std::vector<std::vector<struct XsectCmplxTerm> > get_cmplx_terms(
    const std::string & filename);
std::vector<std::vector<struct XsectPolyTerm> > get_poly_terms(
    const std::string & filename);

// xsection pieces stored in groups of 4 ints: prefactor, ti, cexpr, pexpr.
std::vector<int> get_xsect_reg_pieces(const std::string & filename);
std::vector<int> get_xsect_ti_pieces(const std::string & filename);

std::vector<std::complex<double> > calc_isme_sums(
    const std::vector< std::vector< std::pair<int, int> > > & isme_sums,
    std::vector<std::complex<double> > & isme);

std::vector<Poly3> calc_isme_sums(
    const std::vector< std::vector< std::pair<int, int> > > & isme_sums,
    std::vector<Poly3> & isme);

void calc_cmplx_terms(
    const std::vector<std::vector<struct XsectCmplxTerm> > & termlist,
    const std::complex<double> *sssums,
    const std::vector<std::complex<double> > & ismesums,
    std::complex<double> *exprs);
void calc_cmplx_terms(
    const std::vector<std::vector<struct XsectCmplxTerm> > & termlist,
    const std::complex<double> *sssums,
    const std::vector<Poly3> & ismesums,
    Poly3 *exprs);
void calc_poly_terms(
    const std::vector<std::vector<struct XsectPolyTerm> > & termlist,
    const std::complex<double> *sssums,
    const std::vector<std::complex<double> > & ismesums,
    const Poly3 *polys, Poly3 *exprs);
void calc_poly_terms(
    const std::vector<std::vector<struct XsectPolyTerm> > & termlist,
    const std::complex<double> *sssums,
    const std::vector<Poly3> & ismesums,
    const Poly3 *polys, Poly3 *exprs);

Poly3 calc_xsect_reg(const std::vector<int> & pieces,
                     const std::complex<double> *prefactors,
                     const std::complex<double> *exprs, const Poly3 *expr2s);
Poly3 calc_xsect_pole(const std::vector<int> & pieces,
                      const std::complex<double> *prefactors,
                      const std::complex<double> *exprs, const Poly3 *expr2s);
Poly3 calc_xsect_ti(const std::vector<int> & pieces,
                    const std::complex<double> *prefactors,
                    std::vector<CoefficientNode::Ptr> & ti,
                    const std::complex<double> *exprs, const Poly3 *expr2s);
Poly3 calc_xsect_reg(const std::vector<int> & pieces,
                     const std::complex<double> *prefactors,
                     const Poly3 *exprs, const Poly3 *expr2s);
Poly3 calc_xsect_pole(const std::vector<int> & pieces,
                      const std::complex<double> *prefactors,
                      const Poly3 *exprs, const Poly3 *expr2s);
Poly3 calc_xsect_ti(const std::vector<int> & pieces,
                    const std::complex<double> *prefactors,
                    std::vector<CoefficientNode::Ptr> & ti,
                    const Poly3 *exprs, const Poly3 *expr2s);

void populate_tis_by_file(TRed *tred,
                          std::vector<CoefficientNode::Ptr> & ti,
                          const std::string & filename);

#endif // XSECTION_H_
