#ifndef AB_NODE_H_
#define AB_NODE_H_

#include "evaluation.h"
#include "tcomplex.h"

/**********************************************************************
 *                                                                    *
 * ANode                                                              *
 *                                                                    *
 **********************************************************************/

class ANode : public EvaluationNode {
  public:
    typedef shared_ptr<ANode> Ptr;
    ANode(const Coefficient & coeff, const CoefficientNode & cnode, 
          EvaluationList* parent, CoefficientPool* pool);

    void add_dependencies() {}
    void evaluate();

  private:
    template <typename fT> void evaluateT(Poly3T<fT> & value);
};

/**********************************************************************
 *                                                                    *
 * BNode                                                              *
 *                                                                    *
 **********************************************************************/

class BNode : public EvaluationNode {
  public:
    // precision before switching B evaluation to on-shell, also used for
    // choosing branches of log functions
    // WARNING: This number should probably depend on the truncation
    // precision of the scalar interface. Smaller values have been
    // found problematic in phase space points checks against the
    // literature.
    static double zerocutoff_;

    typedef shared_ptr<BNode> Ptr;
    BNode(const Coefficient & coeff, const CoefficientNode & cnode,
          EvaluationList* parent, CoefficientPool* pool);
    ~BNode();

    // Adds all dependencies of the current node to the node.
    // NOTE: This has side effects, as it adds to the
    // CoefficientPool
    void add_dependencies();
    void evaluate();

  private:
    template <typename fT> typename tcomplex<fT>::type b1n_kernel(
        int n, typename tcomplex<fT>::type arg);
    template <typename fT> Poly3T<fT> b1n(int n);
    template <typename fT> void evaluateT(Poly3T<fT> & value);

    EvaluationNode::Ptr b_di0_;
    EvaluationNode::Ptr b_di0a1_;
    ContractedFirstProp* b_dp0_;

    // should probably make this dependent on current precision
    static const int sumterms_ = 20; // how many terms in B1 sum to do

    static const double eps_feyn_;
};

#endif  // AB_NODE_H_
