#ifndef DD_NODE_H_
#define DD_NODE_H_

#include <map>
#include "evaluation.h"

/**********************************************************************
 *                                                                    *
 * DDNode                                                             *
 *                                                                    *
 **********************************************************************/

class DDNode : public EvaluationNode {
  public:
    typedef shared_ptr<DDNode> Ptr;
    DDNode(const Coefficient & coeff, const CoefficientNode & cnode,
           EvaluationList* parent, CoefficientPool* pool)
    : EvaluationNode(coeff, cnode, parent, pool), t_dp0_(NULL) {}
    ~DDNode() {
      if (t_dp0_ != NULL)
        delete t_dp0_;
    }

    // Adds all dependencies of the current node to the node.
    // NOTE: This has side effects, as it adds to the
    // CoefficientPool
    void add_dependencies();

    void evaluate();

    bool operator==(const DDNode & node) const;
    bool operator==(const DDNode & node);

    bool cmp(const EvaluationNode & node) const;

  private:
    template <typename fT> void evaluateT(Poly3T<fT> & value_);
    template <typename fT> void get_shat(unsigned int n,
                                         Poly3T<fT> & s_value_);

    const static double s_cayley_zero_thresh_;

    // D with missing first index
    EvaluationNode::Ptr t_di0_;
    // D with dropped first index, 0 prepended
    EvaluationNode::Ptr t_di0_ai0_;
    // D with missing index, not first 0
    std::map<int, EvaluationNode::Ptr> t_dr_;
    // D missing first and another index, 0 prepended
    std::map<int, EvaluationNode::Ptr> t_di0_ai0_dr_;
    // C(k) (dropped middle propagator)
    std::map<int, CoefficientNode::Ptr> t_dpn_;
    // C(0) (dropped first propagator)
    ContractedFirstProp* t_dp0_;
};

#endif  // DD_NODE_H_
