#ifndef ET_NODE_H_
#define ET_NODE_H_

#include <map>
#include <vector>
#include "evaluation.h"
#include "tcomplex.h"

/**********************************************************************
 *                                                                    *
 * ETNode                                                             *
 *                                                                    *
 **********************************************************************/

class ETNode : public EvaluationNode {
  public:
    typedef shared_ptr<ETNode> Ptr;
    ETNode(const Coefficient & coeff, const CoefficientNode & cnode,
           EvaluationList* parent, CoefficientPool* pool)
    : EvaluationNode(coeff, cnode, parent, pool) {}

    // Adds all dependencies of the current node to the node.
    // NOTE: This has side effects, as it adds to the
    // CoefficientPool
    void add_dependencies();
    void evaluate();

  private:
    Coefficient make_scalar();
    template <typename fT> void evaluateT(Poly3T<fT> & value);
    template <typename fT> void scalar_evaluate(Poly3T<fT> & value);
    template <typename fT> void r1_evaluate(Poly3T<fT> & value);
    template <typename fT> void r2_evaluate(Poly3T<fT> & value);
    template <typename fT> void r3_evaluate(Poly3T<fT> & value);

    template <typename fT> typename tcomplex<fT>::type xe(int i, int j,
                                                          int s, int t);

    template <typename fT> typename tcomplex<fT>::type E00i_I4s(int j, int s);
    template <typename fT> typename tcomplex<fT>::type E00i_I3st(int j,
                                                                 int s, int t);
    template <typename fT> typename tcomplex<fT>::type E00i_I2stu(int j,
                                                                  int s,
                                                                  int t,
                                                                  int u);

    template <typename fT> typename tcomplex<fT>::type Eijk_I2stu_part(int i,
                                                                       int j,
                                                                       int k,
                                                                       int s,
                                                                       int t,
                                                                       int u);
    template <typename fT> typename tcomplex<fT>::type Eijk_I2stu(int i,
                                                                  int j,
                                                                  int k,
                                                                  int s,
                                                                  int t,
                                                                  int u);
    template <typename fT> typename tcomplex<fT>::type Eijk_I3st_part(int i,
                                                                      int j,
                                                                      int k,
                                                                      int s,
                                                                      int t);
    template <typename fT> typename tcomplex<fT>::type Eijk_I3st(int i, int j,
                                                                 int k,
                                                                 int s, int t);
    template <typename fT> typename tcomplex<fT>::type Eijk_I4s_part(int i,
                                                                     int j,
                                                                     int k,
                                                                     int s);
    template <typename fT> typename tcomplex<fT>::type Eijk_I4s(int i, int j,
                                                                int k, int s);

  private:
    // Dependencies
    struct DepDesc {
      DepDesc(const std::vector<int> & s, const CoefficientNode::Ptr & n)
      : sig(s), order(s.size()), node(n) {};
      std::vector<int> sig;
      int order;
      CoefficientNode::Ptr node;
    };
    typedef std::vector<DepDesc> DepVector;
    DepVector ds_i_;
    DepVector cs_ij_;
    DepVector bs_ijk_;
};


/**********************************************************************
 *                                                                    *
 * Inlines                                                            *
 *                                                                    *
 **********************************************************************/

inline Coefficient ETNode::make_scalar() {
  return Coefficient(coeff_.points(),
                     coeff_.invariant_group(),
                     std::vector<int>());
}

#endif  // ET_NODE_H_
