#ifndef FA_NODE_H_
#define FA_NODE_H_

#include <map>
#include <vector>
#include "evaluation.h"
#include "tcomplex.h"

/**********************************************************************
 *                                                                    *
 * FANode                                                             *
 *                                                                    *
 **********************************************************************/

class FANode : public EvaluationNode {
  public:
    typedef shared_ptr<FANode> Ptr;
    FANode(const Coefficient & coeff, const CoefficientNode & cnode,
           EvaluationList* parent, CoefficientPool* pool);
    ~FANode();

    // Adds all dependencies of the current node to the node.
    // NOTE: This has side effects, as it adds to the
    // CoefficientPool
    void add_dependencies();
    void evaluate();

    typedef std::map<int, CoefficientNode::Ptr> IntNodeMap;
    typedef std::map<int, ContractedFirstProp*> IntCFPMap;

  private:
    void add_scalar_dependencies();
    void add_f_bar_dependencies(const std::vector<int> & inds, int f_bar_index);

    template <typename fT> void evaluateT(Poly3T<fT> & value);

    template <typename fT> Poly3T<fT> scalar_evaluate();
    // TODO: remove inds argument, it's already available through
    // member f_bar_indices_
    template <typename fT>
    Poly3T<fT> f_bar_evaluate(const std::vector<int> & inds, int index_id);

    bool coeff_has_permutations();
    // populates f_bar_indices_ with all possible swappings for
    // the evlaution of symmetric E coefficients from F_bar's
    void coeff_index_swappings();

  private:
    // Indices of F_bar's needed to evaluate the value for coeff_.
    // First the index, second the multiplicity of equal
    // configurations. Imutable after construction.
    std::map<std::vector<int>, int> f_bar_indices_;

    // Dependencies
    //   for F_{i1...ip}: d_dp0_ = E_{i2...ip}(0)
    IntCFPMap e_dp0_;

    // for F_{i1...ip}: d_dpn_ = E_{i2...ip}(n)
    std::map<int, IntNodeMap> e_dpn_;
};

#endif  // FA_NODE_H_
