#ifndef FT_NODE_H_
#define FT_NODE_H_

#include <map>
#include <vector>
#include "./evaluation.h"
#include "tcomplex.h"

/**********************************************************************
 *                                                                    *
 * FTNode                                                             *
 *                                                                    *
 **********************************************************************/

class FTNode : public EvaluationNode {
  public:
    typedef shared_ptr<FTNode> Ptr;
    FTNode(const Coefficient & coeff, const CoefficientNode & cnode,
           EvaluationList* parent, CoefficientPool* pool)
    : EvaluationNode(coeff, cnode, parent, pool) {}

    // Adds all dependencies of the current node to the node.
    // NOTE: This has side effects, as it adds to the
    // CoefficientPool
    void add_dependencies();
    void evaluate();

  private:
    CoefficientNode::Ptr add_child_coefficient(const Coefficient & coeff,
                                               int drop_prop);
    template <typename fT> void evaluateT(Poly3T<fT> & value);
    template <typename fT> void scalar_evaluate(Poly3T<fT> & value);
    template <typename fT> void r1_evaluate(Poly3T<fT> & value);
    template <typename fT> void r2_evaluate(Poly3T<fT> & value);
    template <typename fT> void r3_evaluate(Poly3T<fT> & value);

    template <typename fT> typename tcomplex<fT>::type vr(int i, int r);

  private:
    typedef std::map<int, CoefficientNode::Ptr> DepMap;
    typedef std::map<int, ContractedFirstProp*> CFPDepMap;

    // first: order index, second: map of dropped propagators
    std::map<int, DepMap> e_dpn_;
    CFPDepMap e_dp0_;
};

#endif  // FT_NODE_H_
