#ifndef POLY3_H_
#define POLY3_H_

#include <iostream>
#include <sstream>
#include <string>
#include "tcomplex.h"


/**********************************************************************
 *                                                                    *
 * Poly3                                                              *
 *                                                                    *
 **********************************************************************/
/*
 * WARNING: Poly3 has side effects, though most removed.
 */

template <typename fT> class Poly3T {
  typedef typename tcomplex<fT>::type cfT;

  public:
    explicit Poly3T<fT>(int minorder=-2):_minorder(minorder) {
      // Note: we should probably check minorder > -2, < 0
      _coefficient[0] = cfT(0, 0);
      _coefficient[1] = cfT(0, 0);
      _coefficient[2] = cfT(0, 0);
    }

    Poly3T<fT>(const Poly3T<fT> & cv) {
      for (int i = 0; i < 3; ++i)
        _coefficient[i] = cv._coefficient[i];
      _minorder = cv._minorder;
    }

    template <typename T2> Poly3T<fT>(const Poly3T<T2> & cv) {
      for (int i = 0; i < 3; ++i) {
        fT x = fT(cv._coefficient[i].real());
        fT y = fT(cv._coefficient[i].imag());
        _coefficient[i] = cfT(x, y);
      }
      _minorder = cv._minorder;
    }

    Poly3T<fT>(cfT x0, cfT x1, cfT x2, int minorder=-2) : _minorder(minorder) {
      _coefficient[0] = x0;
      _coefficient[1] = x1;
      _coefficient[2] = x2;
    }

    Poly3T<fT>(fT x, int minorder = -2) : _minorder(minorder) {
      _coefficient[0] = 0;
      _coefficient[1] = 0;
      _coefficient[2] = cfT(x, 0);
    }

    Poly3T<fT>(cfT x, int minorder = -2):_minorder(minorder) {
      _coefficient[0] = 0;
      _coefficient[1] = 0;
      _coefficient[2] = x;
    }

    const cfT order(int i) const {
      return _coefficient[i-_minorder];
    }

    void print(std::string prefix, int digits = 16,
               std::ostream & strm = std::cout) const;

    // operators

    Poly3T<fT>& operator=(const Poly3T<fT> & cv);
    Poly3T<fT>& operator+=(const Poly3T<fT> &cv);
    Poly3T<fT>& operator-=(const Poly3T<fT> &cv);

    template <typename T2> Poly3T<fT>& operator*=(const T2 & val);

    Poly3T<fT>& operator*=(const Poly3T<fT> & poly);
//       Poly3T<T>& operator/=(const T & val);

    Poly3T<fT> operator-() const; // added by Seth, probably temporary

    cfT _coefficient[3];
    int _minorder;
};

// Template functions

template <typename fT> template <typename T2> inline Poly3T<fT> &
Poly3T<fT>::operator*=(const T2 & val) {
  _coefficient[0] *= val;
  _coefficient[1] *= val;
  _coefficient[2] *= val;
  return *this;
}

template <typename fT> inline Poly3T<fT> & Poly3T<fT>::operator*=(
    const Poly3T<fT> & poly) {
  if (poly._minorder - _minorder > 2) {
    std::cerr << "Trying to multiply indices of wrong orders" << std::endl;
    print("e");
    poly.print("e");
    exit(1);
  }
  typename tcomplex<fT>::type tmp[3];
  tmp[0] = _coefficient[0]*poly._coefficient[0];

  tmp[1] = _coefficient[0]*poly._coefficient[1]
         + _coefficient[1]*poly._coefficient[0];

  tmp[2] = _coefficient[0]*poly._coefficient[2]
         + _coefficient[1]*poly._coefficient[1]
         + _coefficient[2]*poly._coefficient[0];

  for (int i = 0; i < 3; ++i)
    _coefficient[i] = tmp[i];

  return *this;
}

template <typename fT> inline Poly3T<fT> & Poly3T<fT>::operator=(
    const Poly3T<fT> & cv) {
  if (this == &cv)
    return *this;

  for (int i = 0; i < 3; ++i)
    _coefficient[i] = cv._coefficient[i];

  _minorder = cv._minorder;
  return *this;
}

// TODO: Runtime checks below probably take a long time, consider removing to debug.
template <typename fT> inline Poly3T<fT> & Poly3T<fT>::operator+=(
    const Poly3T<fT> &cv) {
  if (_minorder != cv._minorder) {
    std::cerr << "Attempting to add different orders of Poly3" << std::endl;
    this->print("ep");
    cv.print("ep");
    abort();
    //exit(1);
  }

  _coefficient[0] += cv._coefficient[0];
  _coefficient[1] += cv._coefficient[1];
  _coefficient[2] += cv._coefficient[2];
  return *this;
}

template <typename fT> inline Poly3T<fT> & Poly3T<fT>::operator-=(
    const Poly3T<fT> &cv) {
  if (_minorder != cv._minorder) {
    std::cerr << "Attempting to add different orders of Poly3" << std::endl;
    this->print("ep");
    cv.print("ep");
    abort();
//    exit(1);
  }

  _coefficient[0] -= cv._coefficient[0];
  _coefficient[1] -= cv._coefficient[1];
  _coefficient[2] -= cv._coefficient[2];
  return *this;
}

template <typename fT> inline Poly3T<fT> Poly3T<fT>::operator-() const {
  return Poly3T<fT>(_minorder) -= *this;
}

template <typename fT> inline Poly3T<fT> operator*(Poly3T<fT> lhs,
                                                   const Poly3T<fT> &cv) {
  Poly3T<fT> res(lhs);
  return res *= cv;
}

template <typename fT> inline Poly3T<fT> operator+(Poly3T<fT> lhs,
                                                   const Poly3T<fT> &cv) {
  Poly3T<fT> res(lhs);
  return res += cv;
}

// Note: we access _coefficient here, which should probably be private.
// Perhaps define a new Poly3 method that does += on scalars?
// (Applies to +/- functions below.)

template <typename fT, typename T2> inline Poly3T<fT> operator+(
    Poly3T<fT> lhs, const T2 &val) {
  Poly3T<fT> res(lhs);
  res._coefficient[0-res._minorder] += val;
  return res;
}

template <typename fT, typename T2> inline Poly3T<fT> operator+(
    const T2 &val,
    Poly3T<fT> rhs) {
  Poly3T<fT> res(rhs);
  res._coefficient[0-res._minorder] += val;
  return res;
}

template <typename fT> inline Poly3T<fT> operator-(Poly3T<fT> lhs,
                                                   const Poly3T<fT> &cv) {
  Poly3T<fT> res(lhs);
  return res -= cv;
}

template <typename fT, typename T2> inline Poly3T<fT> operator-(
    Poly3T<fT> lhs, const T2 &val) {
  Poly3T<fT> res(lhs);
  res._coefficient[0-res._minorder] -= val;
  return res;
}

template <typename fT, typename T2> inline Poly3T<fT> operator-(
    const T2 &val, Poly3T<fT> rhs) {
  Poly3T<fT> res(-rhs);
  res._coefficient[0-res._minorder] += val;
  return res;
}

template <typename fT, typename T2> inline Poly3T<fT> operator*(
    Poly3T<fT> lhs, const T2 &val) {
  Poly3T<fT> res(lhs);
  return res *= val;
}

template <typename fT, typename T2> inline Poly3T<fT> operator*(
    const T2 & val, Poly3T<fT> rhs) {
  Poly3T<fT> res(rhs);
  return res *= val;
}

template <typename fT> void Poly3T<fT>::print(std::string prefix, int digits,
                                              std::ostream & strm) const {
  strm << std::setprecision(digits);
  strm << prefix << _minorder << "\t" << _coefficient[0]  << std::endl
       << prefix << (_minorder + 1) << "\t" << _coefficient[1] << std::endl
       << prefix << (_minorder + 2) << "\t" << _coefficient[2]  << std::endl;
}

typedef Poly3T<double> Poly3;

#endif  // POLY3_H_
