#ifndef PRECISION_H_
#define PRECISION_H_

// In the future, one should probably define helper classes inside
// the evaluation nodes that handle the math, and these should be templated.
// For now, needed functions are templated. For the case of evaluate(), this
// means that due to the conflict between virtual/template methods, the same
// evaluation code must appear in each node type to decide which type to use.
// This shouldn't be an issue if the internal math objects to EvaluationNode
// were templated instead.
// In gcc on x64 machines, long double is 80-bit.
// TODO: Check that things still work when fT1 != double--should be hard-coded
// doubles in interface calls, fT1 on tred side of things.

/* MAX_PREC_LVL: how many levels of floating point to attempt. We may want to
   define this variable at the makefile level once configuration tools are
   available, as well as the specific precisions below. */
#define MAX_PREC_LVL 3

typedef double fT1; // first attempted precision
typedef long double fT2; // second attempted precision
#if MAX_PREC_LVL >= 3
#include <quadmath.h>
typedef __float128 fT3; // third attempted precision

// Overloads for float128
#if __GNUC__ < 7
// gcc 7 and higher seem to define this overload for some reason.
// Note: to get correct floating abs in template functions, need
// "using namespace std".
inline __float128 abs(__float128 x) { return fabsq(x); }
#endif
inline __float128 fabs(__float128 x) { return fabsq(x); }
inline __float128 pow(__float128 base, __float128 exp) {
  return powq(base, exp);
}
inline __float128 pow(__float128 base, int exp) {
  return powq(base, exp);
}

#endif

#endif // PRECISION_H_
