#ifndef PV_NODE_H_
#define PV_NODE_H_

#include <map>
#include "evaluation.h"

/**********************************************************************
 *                                                                    *
 * PVNode                                                             *
 *                                                                    *
 **********************************************************************/

class PVNode : public EvaluationNode {
  public:
    typedef shared_ptr<PVNode> Ptr;
    PVNode(const Coefficient & coeff, const CoefficientNode & cnode,
           EvaluationList* parent, CoefficientPool* pool)
    : EvaluationNode(coeff, cnode, parent, pool), t_dp0_(NULL) {}
    ~PVNode() {
      if (t_dp0_ != NULL)
        delete t_dp0_;
    }

    // Adds all dependencies of the current node to the node.
    // NOTE: This has side effects, as it adds to the
    // CoefficientPool
    void add_dependencies();

    void evaluate();

    bool operator==(const PVNode & node) const;
    bool operator==(const PVNode & node);

  private:
    template <typename fT> void evaluateT(Poly3T<fT> & value_);

    EvaluationNode::Ptr t_di0_;
    std::map<int, EvaluationNode::Ptr> t_di0_ain_;
    std::map<int, EvaluationNode::Ptr> t_di0_ai0_dr_;
    std::map<int, CoefficientNode::Ptr> t_dpn_;
    ContractedFirstProp* t_dp0_;

    // We might want to make gram_thresh more universal (across eval nodes)
    // than just putting it here in the future
    // Note, there are 3 of these for up to 3 precision levels
    static const double gram_thresh_[3];
};

#endif  // PV_NODE_H_
